<!DOCTYPE html>
<html>
  <head>
    <meta charset="utf-8"/>
    <title><%= htmlWebpackPlugin.options.title %></title>
    <script>
      window.PRET_PICKLE_FILE = '<%= __PRET_PICKLE_FILE__ %>';
      window.PRET_REMOTE_IMPORTS = '__PRET_REMOTE_IMPORTS__';
    </script>
    /* PRET_HEAD_TAGS */
    <%= htmlWebpackPlugin.tags.headTags %>
    <style>
        body, html, #root, .pret-render-container {
            margin: 0;
            padding: 0;
            font-family: sans-serif;
            width: 100%;
            height: 100%;
        }
    </style>
  </head>
  <body>
    <%= htmlWebpackPlugin.tags.bodyTags %>
    <div class='pret-render-container' data-pret-chunk-idx=0></div>
  </body>
</html>
