import sys
from typing import Any, Union

from pret.marshal import make_stub_js_module, js
from pret.render import stub_component

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

__version__ = "0.1.0"

make_stub_js_module("React", "pret", "react", __version__, __name__)

props_mapping = {
    "href_lang": "hrefLang",
    "referrer_policy": "referrerPolicy",
    "default_checked": "defaultChecked",
    "default_value": "defaultValue",
    "suppress_content_editable_warning": "suppressContentEditableWarning",
    "suppress_hydration_warning": "suppressHydrationWarning",
    "access_key": "accessKey",
    "auto_focus": "autoFocus",
    "class_name": "className",
    "content_editable": "contentEditable",
    "context_menu": "contextMenu",
    "spell_check": "spellCheck",
    "tab_index": "tabIndex",
    "radio_group": "radioGroup",
    "auto_capitalize": "autoCapitalize",
    "auto_correct": "autoCorrect",
    "auto_save": "autoSave",
    "item_prop": "itemProp",
    "item_scope": "itemScope",
    "item_type": "itemType",
    "item_id": "itemID",
    "item_ref": "itemRef",
    "input_mode": "inputMode",
    "is_": "is",
    "aria_activedescendant": "aria-activedescendant",
    "aria_atomic": "aria-atomic",
    "aria_autocomplete": "aria-autocomplete",
    "aria_busy": "aria-busy",
    "aria_checked": "aria-checked",
    "aria_colcount": "aria-colcount",
    "aria_colindex": "aria-colindex",
    "aria_colspan": "aria-colspan",
    "aria_controls": "aria-controls",
    "aria_current": "aria-current",
    "aria_describedby": "aria-describedby",
    "aria_details": "aria-details",
    "aria_disabled": "aria-disabled",
    "aria_dropeffect": "aria-dropeffect",
    "aria_errormessage": "aria-errormessage",
    "aria_expanded": "aria-expanded",
    "aria_flowto": "aria-flowto",
    "aria_grabbed": "aria-grabbed",
    "aria_haspopup": "aria-haspopup",
    "aria_hidden": "aria-hidden",
    "aria_invalid": "aria-invalid",
    "aria_keyshortcuts": "aria-keyshortcuts",
    "aria_label": "aria-label",
    "aria_labelledby": "aria-labelledby",
    "aria_level": "aria-level",
    "aria_live": "aria-live",
    "aria_modal": "aria-modal",
    "aria_multiline": "aria-multiline",
    "aria_multiselectable": "aria-multiselectable",
    "aria_orientation": "aria-orientation",
    "aria_owns": "aria-owns",
    "aria_placeholder": "aria-placeholder",
    "aria_posinset": "aria-posinset",
    "aria_pressed": "aria-pressed",
    "aria_readonly": "aria-readonly",
    "aria_relevant": "aria-relevant",
    "aria_required": "aria-required",
    "aria_roledescription": "aria-roledescription",
    "aria_rowcount": "aria-rowcount",
    "aria_rowindex": "aria-rowindex",
    "aria_rowspan": "aria-rowspan",
    "aria_selected": "aria-selected",
    "aria_setsize": "aria-setsize",
    "aria_sort": "aria-sort",
    "aria_valuemax": "aria-valuemax",
    "aria_valuemin": "aria-valuemin",
    "aria_valuenow": "aria-valuenow",
    "aria_valuetext": "aria-valuetext",
    "dangerously_set_inner_html": "dangerouslySetInnerHTML",
    "on_copy": "onCopy",
    "on_copy_capture": "onCopyCapture",
    "on_cut": "onCut",
    "on_cut_capture": "onCutCapture",
    "on_paste": "onPaste",
    "on_paste_capture": "onPasteCapture",
    "on_composition_end": "onCompositionEnd",
    "on_composition_end_capture": "onCompositionEndCapture",
    "on_composition_start": "onCompositionStart",
    "on_composition_start_capture": "onCompositionStartCapture",
    "on_composition_update": "onCompositionUpdate",
    "on_composition_update_capture": "onCompositionUpdateCapture",
    "on_focus": "onFocus",
    "on_focus_capture": "onFocusCapture",
    "on_blur": "onBlur",
    "on_blur_capture": "onBlurCapture",
    "on_change": "onChange",
    "on_change_capture": "onChangeCapture",
    "on_before_input": "onBeforeInput",
    "on_before_input_capture": "onBeforeInputCapture",
    "on_input": "onInput",
    "on_input_capture": "onInputCapture",
    "on_reset": "onReset",
    "on_reset_capture": "onResetCapture",
    "on_submit": "onSubmit",
    "on_submit_capture": "onSubmitCapture",
    "on_invalid": "onInvalid",
    "on_invalid_capture": "onInvalidCapture",
    "on_load": "onLoad",
    "on_load_capture": "onLoadCapture",
    "on_error": "onError",
    "on_error_capture": "onErrorCapture",
    "on_key_down": "onKeyDown",
    "on_key_down_capture": "onKeyDownCapture",
    "on_key_press": "onKeyPress",
    "on_key_press_capture": "onKeyPressCapture",
    "on_key_up": "onKeyUp",
    "on_key_up_capture": "onKeyUpCapture",
    "on_abort": "onAbort",
    "on_abort_capture": "onAbortCapture",
    "on_can_play": "onCanPlay",
    "on_can_play_capture": "onCanPlayCapture",
    "on_can_play_through": "onCanPlayThrough",
    "on_can_play_through_capture": "onCanPlayThroughCapture",
    "on_duration_change": "onDurationChange",
    "on_duration_change_capture": "onDurationChangeCapture",
    "on_emptied": "onEmptied",
    "on_emptied_capture": "onEmptiedCapture",
    "on_encrypted": "onEncrypted",
    "on_encrypted_capture": "onEncryptedCapture",
    "on_ended": "onEnded",
    "on_ended_capture": "onEndedCapture",
    "on_loaded_data": "onLoadedData",
    "on_loaded_data_capture": "onLoadedDataCapture",
    "on_loaded_metadata": "onLoadedMetadata",
    "on_loaded_metadata_capture": "onLoadedMetadataCapture",
    "on_load_start": "onLoadStart",
    "on_load_start_capture": "onLoadStartCapture",
    "on_pause": "onPause",
    "on_pause_capture": "onPauseCapture",
    "on_play": "onPlay",
    "on_play_capture": "onPlayCapture",
    "on_playing": "onPlaying",
    "on_playing_capture": "onPlayingCapture",
    "on_progress": "onProgress",
    "on_progress_capture": "onProgressCapture",
    "on_rate_change": "onRateChange",
    "on_rate_change_capture": "onRateChangeCapture",
    "on_resize": "onResize",
    "on_resize_capture": "onResizeCapture",
    "on_seeked": "onSeeked",
    "on_seeked_capture": "onSeekedCapture",
    "on_seeking": "onSeeking",
    "on_seeking_capture": "onSeekingCapture",
    "on_stalled": "onStalled",
    "on_stalled_capture": "onStalledCapture",
    "on_suspend": "onSuspend",
    "on_suspend_capture": "onSuspendCapture",
    "on_time_update": "onTimeUpdate",
    "on_time_update_capture": "onTimeUpdateCapture",
    "on_volume_change": "onVolumeChange",
    "on_volume_change_capture": "onVolumeChangeCapture",
    "on_waiting": "onWaiting",
    "on_waiting_capture": "onWaitingCapture",
    "on_aux_click": "onAuxClick",
    "on_aux_click_capture": "onAuxClickCapture",
    "on_click": "onClick",
    "on_click_capture": "onClickCapture",
    "on_context_menu": "onContextMenu",
    "on_context_menu_capture": "onContextMenuCapture",
    "on_double_click": "onDoubleClick",
    "on_double_click_capture": "onDoubleClickCapture",
    "on_drag": "onDrag",
    "on_drag_capture": "onDragCapture",
    "on_drag_end": "onDragEnd",
    "on_drag_end_capture": "onDragEndCapture",
    "on_drag_enter": "onDragEnter",
    "on_drag_enter_capture": "onDragEnterCapture",
    "on_drag_exit": "onDragExit",
    "on_drag_exit_capture": "onDragExitCapture",
    "on_drag_leave": "onDragLeave",
    "on_drag_leave_capture": "onDragLeaveCapture",
    "on_drag_over": "onDragOver",
    "on_drag_over_capture": "onDragOverCapture",
    "on_drag_start": "onDragStart",
    "on_drag_start_capture": "onDragStartCapture",
    "on_drop": "onDrop",
    "on_drop_capture": "onDropCapture",
    "on_mouse_down": "onMouseDown",
    "on_mouse_down_capture": "onMouseDownCapture",
    "on_mouse_enter": "onMouseEnter",
    "on_mouse_leave": "onMouseLeave",
    "on_mouse_move": "onMouseMove",
    "on_mouse_move_capture": "onMouseMoveCapture",
    "on_mouse_out": "onMouseOut",
    "on_mouse_out_capture": "onMouseOutCapture",
    "on_mouse_over": "onMouseOver",
    "on_mouse_over_capture": "onMouseOverCapture",
    "on_mouse_up": "onMouseUp",
    "on_mouse_up_capture": "onMouseUpCapture",
    "on_select": "onSelect",
    "on_select_capture": "onSelectCapture",
    "on_touch_cancel": "onTouchCancel",
    "on_touch_cancel_capture": "onTouchCancelCapture",
    "on_touch_end": "onTouchEnd",
    "on_touch_end_capture": "onTouchEndCapture",
    "on_touch_move": "onTouchMove",
    "on_touch_move_capture": "onTouchMoveCapture",
    "on_touch_start": "onTouchStart",
    "on_touch_start_capture": "onTouchStartCapture",
    "on_pointer_down": "onPointerDown",
    "on_pointer_down_capture": "onPointerDownCapture",
    "on_pointer_move": "onPointerMove",
    "on_pointer_move_capture": "onPointerMoveCapture",
    "on_pointer_up": "onPointerUp",
    "on_pointer_up_capture": "onPointerUpCapture",
    "on_pointer_cancel": "onPointerCancel",
    "on_pointer_cancel_capture": "onPointerCancelCapture",
    "on_pointer_enter": "onPointerEnter",
    "on_pointer_enter_capture": "onPointerEnterCapture",
    "on_pointer_leave": "onPointerLeave",
    "on_pointer_leave_capture": "onPointerLeaveCapture",
    "on_pointer_over": "onPointerOver",
    "on_pointer_over_capture": "onPointerOverCapture",
    "on_pointer_out": "onPointerOut",
    "on_pointer_out_capture": "onPointerOutCapture",
    "on_got_pointer_capture": "onGotPointerCapture",
    "on_got_pointer_capture_capture": "onGotPointerCaptureCapture",
    "on_lost_pointer_capture": "onLostPointerCapture",
    "on_lost_pointer_capture_capture": "onLostPointerCaptureCapture",
    "on_scroll": "onScroll",
    "on_scroll_capture": "onScrollCapture",
    "on_wheel": "onWheel",
    "on_wheel_capture": "onWheelCapture",
    "on_animation_start": "onAnimationStart",
    "on_animation_start_capture": "onAnimationStartCapture",
    "on_animation_end": "onAnimationEnd",
    "on_animation_end_capture": "onAnimationEndCapture",
    "on_animation_iteration": "onAnimationIteration",
    "on_animation_iteration_capture": "onAnimationIterationCapture",
    "on_transition_end": "onTransitionEnd",
    "on_transition_end_capture": "onTransitionEndCapture",
    "auto_play": "autoPlay",
    "controls_list": "controlsList",
    "cross_origin": "crossOrigin",
    "media_group": "mediaGroup",
    "plays_inline": "playsInline",
    "form_action": "formAction",
    "form_enc_type": "formEncType",
    "form_method": "formMethod",
    "form_no_validate": "formNoValidate",
    "form_target": "formTarget",
    "date_time": "dateTime",
    "on_toggle": "onToggle",
    "on_cancel": "onCancel",
    "on_close": "onClose",
    "accept_charset": "acceptCharset",
    "auto_complete": "autoComplete",
    "enc_type": "encType",
    "no_validate": "noValidate",
    "allow_full_screen": "allowFullScreen",
    "allow_transparency": "allowTransparency",
    "frame_border": "frameBorder",
    "margin_height": "marginHeight",
    "margin_width": "marginWidth",
    "src_doc": "srcDoc",
    "src_set": "srcSet",
    "use_map": "useMap",
    "enter_key_hint": "enterKeyHint",
    "max_length": "maxLength",
    "min_length": "minLength",
    "read_only": "readOnly",
    "key_type": "keyType",
    "key_params": "keyParams",
    "html_for": "htmlFor",
    "as_": "as",
    "image_src_set": "imageSrcSet",
    "image_sizes": "imageSizes",
    "char_set": "charSet",
    "http_equiv": "httpEquiv",
    "class_id": "classID",
    "async_": "async",
    "no_module": "noModule",
    "cell_padding": "cellPadding",
    "cell_spacing": "cellSpacing",
    "col_span": "colSpan",
    "row_span": "rowSpan",
    "dir_name": "dirName",
    "src_lang": "srcLang",
    "disable_picture_in_picture": "disablePictureInPicture",
    "disable_remote_playback": "disableRemotePlayback",
}


@stub_component(js.React.Fragment, props_mapping)
def Fragment(*children): ...


@stub_component(js.React.Suspense, props_mapping)
def Suspense(*children, fallback: Any): ...


@stub_component("a", props_mapping)
def a(
    *children,
    download: Any,
    href: str,
    href_lang: str,
    media: str,
    ping: str,
    target: Union[Any, Literal["_self", "_blank", "_parent", "_top"]],
    type: str,
    referrer_policy: Literal[
        "",
        "no-referrer",
        "no-referrer-when-downgrade",
        "origin",
        "origin-when-cross-origin",
        "same-origin",
        "strict-origin",
        "strict-origin-when-cross-origin",
        "unsafe-url",
    ],
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("abbr", props_mapping)
def abbr(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("address", props_mapping)
def address(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("area", props_mapping)
def area(
    *children,
    alt: str,
    coords: str,
    download: Any,
    href: str,
    href_lang: str,
    media: str,
    referrer_policy: Literal[
        "",
        "no-referrer",
        "no-referrer-when-downgrade",
        "origin",
        "origin-when-cross-origin",
        "same-origin",
        "strict-origin",
        "strict-origin-when-cross-origin",
        "unsafe-url",
    ],
    shape: str,
    target: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("article", props_mapping)
def article(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("aside", props_mapping)
def aside(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("audio", props_mapping)
def audio(
    *children,
    auto_play: bool,
    controls: bool,
    controls_list: str,
    cross_origin: Literal["", "anonymous", "use-credentials"],
    loop: bool,
    media_group: str,
    muted: bool,
    plays_inline: bool,
    preload: str,
    src: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("b", props_mapping)
def b(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("base", props_mapping)
def base(
    *children,
    href: str,
    target: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("bdi", props_mapping)
def bdi(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("bdo", props_mapping)
def bdo(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("big", props_mapping)
def big(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("blockquote", props_mapping)
def blockquote(
    *children,
    cite: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("body", props_mapping)
def body(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("br", props_mapping)
def br(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("button", props_mapping)
def button(
    *children,
    disabled: bool,
    form: str,
    form_action: str,
    form_enc_type: str,
    form_method: str,
    form_no_validate: bool,
    form_target: str,
    name: str,
    type: Literal["button", "submit", "reset"],
    value: Any,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("canvas", props_mapping)
def canvas(
    *children,
    height: Any,
    width: Any,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("caption", props_mapping)
def caption(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("center", props_mapping)
def center(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("cite", props_mapping)
def cite(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("code", props_mapping)
def code(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("col", props_mapping)
def col(
    *children,
    span: Union[int, float],
    width: Any,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("colgroup", props_mapping)
def colgroup(
    *children,
    span: Union[int, float],
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("data", props_mapping)
def data(
    *children,
    value: Any,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("datalist", props_mapping)
def datalist(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("dd", props_mapping)
def dd(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("del", props_mapping)
def del_(
    *children,
    cite: str,
    date_time: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("details", props_mapping)
def details(
    *children,
    open: bool,
    on_toggle: Any,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("dfn", props_mapping)
def dfn(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("dialog", props_mapping)
def dialog(
    *children,
    on_cancel: Any,
    on_close: Any,
    open: bool,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("div", props_mapping)
def div(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("dl", props_mapping)
def dl(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("dt", props_mapping)
def dt(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("em", props_mapping)
def em(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("embed", props_mapping)
def embed(
    *children,
    height: Any,
    src: str,
    type: str,
    width: Any,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("fieldset", props_mapping)
def fieldset(
    *children,
    disabled: bool,
    form: str,
    name: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("figcaption", props_mapping)
def figcaption(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("figure", props_mapping)
def figure(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("footer", props_mapping)
def footer(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("form", props_mapping)
def form(
    *children,
    accept_charset: str,
    action: str,
    auto_complete: str,
    enc_type: str,
    method: str,
    name: str,
    no_validate: bool,
    target: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("h1", props_mapping)
def h1(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("h2", props_mapping)
def h2(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("h3", props_mapping)
def h3(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("h4", props_mapping)
def h4(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("h5", props_mapping)
def h5(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("h6", props_mapping)
def h6(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("head", props_mapping)
def head(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("header", props_mapping)
def header(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("hgroup", props_mapping)
def hgroup(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("hr", props_mapping)
def hr(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("html", props_mapping)
def html(
    *children,
    manifest: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("i", props_mapping)
def i(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("iframe", props_mapping)
def iframe(
    *children,
    allow: str,
    allow_full_screen: bool,
    allow_transparency: bool,
    frame_border: Any,
    height: Any,
    loading: Literal["eager", "lazy"],
    margin_height: Union[int, float],
    margin_width: Union[int, float],
    name: str,
    referrer_policy: Literal[
        "",
        "no-referrer",
        "no-referrer-when-downgrade",
        "origin",
        "origin-when-cross-origin",
        "same-origin",
        "strict-origin",
        "strict-origin-when-cross-origin",
        "unsafe-url",
    ],
    sandbox: str,
    scrolling: str,
    seamless: bool,
    src: str,
    src_doc: str,
    width: Any,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("img", props_mapping)
def img(
    *children,
    alt: str,
    cross_origin: Literal["", "anonymous", "use-credentials"],
    decoding: Literal["async", "auto", "sync"],
    height: Any,
    loading: Literal["eager", "lazy"],
    referrer_policy: Literal[
        "",
        "no-referrer",
        "no-referrer-when-downgrade",
        "origin",
        "origin-when-cross-origin",
        "same-origin",
        "strict-origin",
        "strict-origin-when-cross-origin",
        "unsafe-url",
    ],
    sizes: str,
    src: str,
    src_set: str,
    use_map: str,
    width: Any,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("input", props_mapping)
def input(
    *children,
    accept: str,
    alt: str,
    auto_complete: str,
    capture: Literal[False, True, "user", "environment"],
    checked: bool,
    cross_origin: Literal["", "anonymous", "use-credentials"],
    disabled: bool,
    enter_key_hint: Literal[
        "search", "enter", "done", "go", "next", "previous", "send"
    ],
    form: str,
    form_action: str,
    form_enc_type: str,
    form_method: str,
    form_no_validate: bool,
    form_target: str,
    height: Any,
    list: str,
    max: Any,
    max_length: Union[int, float],
    min: Any,
    min_length: Union[int, float],
    multiple: bool,
    name: str,
    pattern: str,
    placeholder: str,
    read_only: bool,
    required: bool,
    size: Union[int, float],
    src: str,
    step: Any,
    type: Union[
        Any,
        Literal[
            "number",
            "button",
            "checkbox",
            "radio",
            "search",
            "text",
            "tel",
            "url",
            "email",
            "date",
            "time",
            "submit",
            "reset",
            "color",
            "datetime-local",
            "file",
            "hidden",
            "image",
            "month",
            "password",
            "range",
            "week",
        ],
    ],
    value: Any,
    width: Any,
    on_change: Any,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("ins", props_mapping)
def ins(
    *children,
    cite: str,
    date_time: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("kbd", props_mapping)
def kbd(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("keygen", props_mapping)
def keygen(
    *children,
    challenge: str,
    disabled: bool,
    form: str,
    key_type: str,
    key_params: str,
    name: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("label", props_mapping)
def label(
    *children,
    form: str,
    html_for: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("legend", props_mapping)
def legend(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("li", props_mapping)
def li(
    *children,
    value: Any,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("link", props_mapping)
def link(
    *children,
    as_: str,
    cross_origin: Literal["", "anonymous", "use-credentials"],
    href: str,
    href_lang: str,
    integrity: str,
    media: str,
    image_src_set: str,
    image_sizes: str,
    referrer_policy: Literal[
        "",
        "no-referrer",
        "no-referrer-when-downgrade",
        "origin",
        "origin-when-cross-origin",
        "same-origin",
        "strict-origin",
        "strict-origin-when-cross-origin",
        "unsafe-url",
    ],
    sizes: str,
    type: str,
    char_set: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("main", props_mapping)
def main(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("map", props_mapping)
def map(
    *children,
    name: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("mark", props_mapping)
def mark(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("menu", props_mapping)
def menu(
    *children,
    type: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("menuitem", props_mapping)
def menuitem(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("meta", props_mapping)
def meta(
    *children,
    char_set: str,
    http_equiv: str,
    name: str,
    media: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("meter", props_mapping)
def meter(
    *children,
    form: str,
    high: Union[int, float],
    low: Union[int, float],
    max: Any,
    min: Any,
    optimum: Union[int, float],
    value: Any,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("nav", props_mapping)
def nav(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("noscript", props_mapping)
def noscript(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("object", props_mapping)
def object(
    *children,
    class_id: str,
    data: str,
    form: str,
    height: Any,
    name: str,
    type: str,
    use_map: str,
    width: Any,
    wmode: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("ol", props_mapping)
def ol(
    *children,
    reversed: bool,
    start: Union[int, float],
    type: Literal["1", "a", "A", "i", "I"],
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("optgroup", props_mapping)
def optgroup(
    *children,
    disabled: bool,
    label: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("option", props_mapping)
def option(
    *children,
    disabled: bool,
    label: str,
    selected: bool,
    value: Any,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("output", props_mapping)
def output(
    *children,
    form: str,
    html_for: str,
    name: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("p", props_mapping)
def p(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("param", props_mapping)
def param(
    *children,
    name: str,
    value: Any,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("picture", props_mapping)
def picture(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("pre", props_mapping)
def pre(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("progress", props_mapping)
def progress(
    *children,
    max: Any,
    value: Any,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("q", props_mapping)
def q(
    *children,
    cite: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("rp", props_mapping)
def rp(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("rt", props_mapping)
def rt(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("ruby", props_mapping)
def ruby(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("s", props_mapping)
def s(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("samp", props_mapping)
def samp(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("slot", props_mapping)
def slot(
    *children,
    name: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("script", props_mapping)
def script(
    *children,
    async_: bool,
    char_set: str,
    cross_origin: Literal["", "anonymous", "use-credentials"],
    defer: bool,
    integrity: str,
    no_module: bool,
    referrer_policy: Literal[
        "",
        "no-referrer",
        "no-referrer-when-downgrade",
        "origin",
        "origin-when-cross-origin",
        "same-origin",
        "strict-origin",
        "strict-origin-when-cross-origin",
        "unsafe-url",
    ],
    src: str,
    type: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("section", props_mapping)
def section(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("select", props_mapping)
def select(
    *children,
    auto_complete: str,
    disabled: bool,
    form: str,
    multiple: bool,
    name: str,
    required: bool,
    size: Union[int, float],
    value: Any,
    on_change: Any,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("small", props_mapping)
def small(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("source", props_mapping)
def source(
    *children,
    height: Any,
    media: str,
    sizes: str,
    src: str,
    src_set: str,
    type: str,
    width: Any,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("span", props_mapping)
def span(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("strong", props_mapping)
def strong(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("style", props_mapping)
def style(
    *children,
    media: str,
    scoped: bool,
    type: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("sub", props_mapping)
def sub(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("summary", props_mapping)
def summary(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("sup", props_mapping)
def sup(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("table", props_mapping)
def table(
    *children,
    align: Literal["left", "center", "right"],
    bgcolor: str,
    border: Union[int, float],
    cell_padding: Any,
    cell_spacing: Any,
    frame: bool,
    rules: Literal["none", "all", "groups", "rows", "columns"],
    summary: str,
    width: Any,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("template", props_mapping)
def template(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("tbody", props_mapping)
def tbody(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("td", props_mapping)
def td(
    *children,
    align: Literal["left", "center", "right", "justify", "char"],
    col_span: Union[int, float],
    headers: str,
    row_span: Union[int, float],
    scope: str,
    abbr: str,
    height: Any,
    width: Any,
    valign: Literal["top", "middle", "bottom", "baseline"],
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("textarea", props_mapping)
def textarea(
    *children,
    auto_complete: str,
    cols: Union[int, float],
    dir_name: str,
    disabled: bool,
    form: str,
    max_length: Union[int, float],
    min_length: Union[int, float],
    name: str,
    placeholder: str,
    read_only: bool,
    required: bool,
    rows: Union[int, float],
    value: Any,
    wrap: str,
    on_change: Any,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("tfoot", props_mapping)
def tfoot(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("th", props_mapping)
def th(
    *children,
    align: Literal["left", "center", "right", "justify", "char"],
    col_span: Union[int, float],
    headers: str,
    row_span: Union[int, float],
    scope: str,
    abbr: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("thead", props_mapping)
def thead(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("time", props_mapping)
def time(
    *children,
    date_time: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("title", props_mapping)
def title(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("tr", props_mapping)
def tr(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("track", props_mapping)
def track(
    *children,
    default: bool,
    kind: str,
    label: str,
    src: str,
    src_lang: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("u", props_mapping)
def u(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("ul", props_mapping)
def ul(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("var", props_mapping)
def var(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("video", props_mapping)
def video(
    *children,
    height: Any,
    plays_inline: bool,
    poster: str,
    width: Any,
    disable_picture_in_picture: bool,
    disable_remote_playback: bool,
    auto_play: bool,
    controls: bool,
    controls_list: str,
    cross_origin: Literal["", "anonymous", "use-credentials"],
    loop: bool,
    media_group: str,
    muted: bool,
    preload: str,
    src: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("wbr", props_mapping)
def wbr(
    *children,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...


@stub_component("webview", props_mapping)
def webview(
    *children,
    allow_full_screen: bool,
    allowpopups: bool,
    autosize: bool,
    blinkfeatures: str,
    disableblinkfeatures: str,
    disableguestresize: bool,
    disablewebsecurity: bool,
    guestinstance: str,
    httpreferrer: str,
    nodeintegration: bool,
    partition: str,
    plugins: bool,
    preload: str,
    src: str,
    useragent: str,
    webpreferences: str,
    default_checked: bool,
    default_value: Any,
    suppress_content_editable_warning: bool,
    suppress_hydration_warning: bool,
    access_key: str,
    auto_focus: bool,
    class_name: str,
    content_editable: Literal[False, True, "true", "false", "inherit"],
    context_menu: str,
    dir: str,
    draggable: Literal[False, True, "true", "false"],
    hidden: bool,
    id: str,
    lang: str,
    nonce: str,
    placeholder: str,
    slot: str,
    spell_check: Literal[False, True, "true", "false"],
    style: Any,
    tab_index: Union[int, float],
    title: str,
    translate: Literal["yes", "no"],
    radio_group: str,
    role: Union[
        Any,
        Literal[
            "alert",
            "alertdialog",
            "application",
            "article",
            "banner",
            "button",
            "cell",
            "checkbox",
            "columnheader",
            "combobox",
            "complementary",
            "contentinfo",
            "definition",
            "dialog",
            "directory",
            "document",
            "feed",
            "figure",
            "form",
            "grid",
            "gridcell",
            "group",
            "heading",
            "img",
            "link",
            "list",
            "listbox",
            "listitem",
            "log",
            "main",
            "marquee",
            "math",
            "menu",
            "menubar",
            "menuitem",
            "menuitemcheckbox",
            "menuitemradio",
            "navigation",
            "none",
            "note",
            "option",
            "presentation",
            "progressbar",
            "radio",
            "radiogroup",
            "region",
            "row",
            "rowgroup",
            "rowheader",
            "scrollbar",
            "search",
            "searchbox",
            "separator",
            "slider",
            "spinbutton",
            "status",
            "switch",
            "tab",
            "table",
            "tablist",
            "tabpanel",
            "term",
            "textbox",
            "timer",
            "toolbar",
            "tooltip",
            "tree",
            "treegrid",
            "treeitem",
        ],
    ],
    about: str,
    content: str,
    datatype: str,
    inlist: Any,
    prefix: str,
    property: str,
    rel: str,
    resource: str,
    rev: str,
    typeof: str,
    vocab: str,
    auto_capitalize: str,
    auto_correct: str,
    auto_save: str,
    color: str,
    item_prop: str,
    item_scope: bool,
    item_type: str,
    item_id: str,
    item_ref: str,
    results: Union[int, float],
    security: str,
    unselectable: Literal["on", "off"],
    input_mode: Literal[
        "none", "search", "text", "tel", "url", "email", "numeric", "decimal"
    ],
    is_: str,
    aria_activedescendant: str,
    aria_atomic: Literal[False, True, "true", "false"],
    aria_autocomplete: Literal["list", "none", "inline", "both"],
    aria_busy: Literal[False, True, "true", "false"],
    aria_checked: Literal[False, True, "true", "false", "mixed"],
    aria_colcount: Union[int, float],
    aria_colindex: Union[int, float],
    aria_colspan: Union[int, float],
    aria_controls: str,
    aria_current: Literal[
        False, True, "true", "false", "page", "step", "location", "date", "time"
    ],
    aria_describedby: str,
    aria_details: str,
    aria_disabled: Literal[False, True, "true", "false"],
    aria_dropeffect: Literal["link", "none", "copy", "execute", "move", "popup"],
    aria_errormessage: str,
    aria_expanded: Literal[False, True, "true", "false"],
    aria_flowto: str,
    aria_grabbed: Literal[False, True, "true", "false"],
    aria_haspopup: Literal[
        False, True, "true", "false", "dialog", "grid", "listbox", "menu", "tree"
    ],
    aria_hidden: Literal[False, True, "true", "false"],
    aria_invalid: Literal[False, True, "true", "false", "grammar", "spelling"],
    aria_keyshortcuts: str,
    aria_label: str,
    aria_labelledby: str,
    aria_level: Union[int, float],
    aria_live: Literal["off", "assertive", "polite"],
    aria_modal: Literal[False, True, "true", "false"],
    aria_multiline: Literal[False, True, "true", "false"],
    aria_multiselectable: Literal[False, True, "true", "false"],
    aria_orientation: Literal["horizontal", "vertical"],
    aria_owns: str,
    aria_placeholder: str,
    aria_posinset: Union[int, float],
    aria_pressed: Literal[False, True, "true", "false", "mixed"],
    aria_readonly: Literal[False, True, "true", "false"],
    aria_relevant: Literal[
        "text",
        "additions",
        "additions removals",
        "additions text",
        "all",
        "removals",
        "removals additions",
        "removals text",
        "text additions",
        "text removals",
    ],
    aria_required: Literal[False, True, "true", "false"],
    aria_roledescription: str,
    aria_rowcount: Union[int, float],
    aria_rowindex: Union[int, float],
    aria_rowspan: Union[int, float],
    aria_selected: Literal[False, True, "true", "false"],
    aria_setsize: Union[int, float],
    aria_sort: Literal["none", "ascending", "descending", "other"],
    aria_valuemax: Union[int, float],
    aria_valuemin: Union[int, float],
    aria_valuenow: Union[int, float],
    aria_valuetext: str,
    dangerously_set_inner_html: Any,
    on_copy: Any,
    on_copy_capture: Any,
    on_cut: Any,
    on_cut_capture: Any,
    on_paste: Any,
    on_paste_capture: Any,
    on_composition_end: Any,
    on_composition_end_capture: Any,
    on_composition_start: Any,
    on_composition_start_capture: Any,
    on_composition_update: Any,
    on_composition_update_capture: Any,
    on_focus: Any,
    on_focus_capture: Any,
    on_blur: Any,
    on_blur_capture: Any,
    on_change: Any,
    on_change_capture: Any,
    on_before_input: Any,
    on_before_input_capture: Any,
    on_input: Any,
    on_input_capture: Any,
    on_reset: Any,
    on_reset_capture: Any,
    on_submit: Any,
    on_submit_capture: Any,
    on_invalid: Any,
    on_invalid_capture: Any,
    on_load: Any,
    on_load_capture: Any,
    on_error: Any,
    on_error_capture: Any,
    on_key_down: Any,
    on_key_down_capture: Any,
    on_key_press: Any,
    on_key_press_capture: Any,
    on_key_up: Any,
    on_key_up_capture: Any,
    on_abort: Any,
    on_abort_capture: Any,
    on_can_play: Any,
    on_can_play_capture: Any,
    on_can_play_through: Any,
    on_can_play_through_capture: Any,
    on_duration_change: Any,
    on_duration_change_capture: Any,
    on_emptied: Any,
    on_emptied_capture: Any,
    on_encrypted: Any,
    on_encrypted_capture: Any,
    on_ended: Any,
    on_ended_capture: Any,
    on_loaded_data: Any,
    on_loaded_data_capture: Any,
    on_loaded_metadata: Any,
    on_loaded_metadata_capture: Any,
    on_load_start: Any,
    on_load_start_capture: Any,
    on_pause: Any,
    on_pause_capture: Any,
    on_play: Any,
    on_play_capture: Any,
    on_playing: Any,
    on_playing_capture: Any,
    on_progress: Any,
    on_progress_capture: Any,
    on_rate_change: Any,
    on_rate_change_capture: Any,
    on_resize: Any,
    on_resize_capture: Any,
    on_seeked: Any,
    on_seeked_capture: Any,
    on_seeking: Any,
    on_seeking_capture: Any,
    on_stalled: Any,
    on_stalled_capture: Any,
    on_suspend: Any,
    on_suspend_capture: Any,
    on_time_update: Any,
    on_time_update_capture: Any,
    on_volume_change: Any,
    on_volume_change_capture: Any,
    on_waiting: Any,
    on_waiting_capture: Any,
    on_aux_click: Any,
    on_aux_click_capture: Any,
    on_click: Any,
    on_click_capture: Any,
    on_context_menu: Any,
    on_context_menu_capture: Any,
    on_double_click: Any,
    on_double_click_capture: Any,
    on_drag: Any,
    on_drag_capture: Any,
    on_drag_end: Any,
    on_drag_end_capture: Any,
    on_drag_enter: Any,
    on_drag_enter_capture: Any,
    on_drag_exit: Any,
    on_drag_exit_capture: Any,
    on_drag_leave: Any,
    on_drag_leave_capture: Any,
    on_drag_over: Any,
    on_drag_over_capture: Any,
    on_drag_start: Any,
    on_drag_start_capture: Any,
    on_drop: Any,
    on_drop_capture: Any,
    on_mouse_down: Any,
    on_mouse_down_capture: Any,
    on_mouse_enter: Any,
    on_mouse_leave: Any,
    on_mouse_move: Any,
    on_mouse_move_capture: Any,
    on_mouse_out: Any,
    on_mouse_out_capture: Any,
    on_mouse_over: Any,
    on_mouse_over_capture: Any,
    on_mouse_up: Any,
    on_mouse_up_capture: Any,
    on_select: Any,
    on_select_capture: Any,
    on_touch_cancel: Any,
    on_touch_cancel_capture: Any,
    on_touch_end: Any,
    on_touch_end_capture: Any,
    on_touch_move: Any,
    on_touch_move_capture: Any,
    on_touch_start: Any,
    on_touch_start_capture: Any,
    on_pointer_down: Any,
    on_pointer_down_capture: Any,
    on_pointer_move: Any,
    on_pointer_move_capture: Any,
    on_pointer_up: Any,
    on_pointer_up_capture: Any,
    on_pointer_cancel: Any,
    on_pointer_cancel_capture: Any,
    on_pointer_enter: Any,
    on_pointer_enter_capture: Any,
    on_pointer_leave: Any,
    on_pointer_leave_capture: Any,
    on_pointer_over: Any,
    on_pointer_over_capture: Any,
    on_pointer_out: Any,
    on_pointer_out_capture: Any,
    on_got_pointer_capture: Any,
    on_got_pointer_capture_capture: Any,
    on_lost_pointer_capture: Any,
    on_lost_pointer_capture_capture: Any,
    on_scroll: Any,
    on_scroll_capture: Any,
    on_wheel: Any,
    on_wheel_capture: Any,
    on_animation_start: Any,
    on_animation_start_capture: Any,
    on_animation_end: Any,
    on_animation_end_capture: Any,
    on_animation_iteration: Any,
    on_animation_iteration_capture: Any,
    on_transition_end: Any,
    on_transition_end_capture: Any,
): ...
